#!/usr/bin/python3
#
# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Calculates information comparing the versions of dependencies in a Rust project
to the versions of dependencies available on Fedora Rawhide.
"""


# isort: STDLIB
import json
import re
import subprocess

# isort: THIRDPARTY
import requests
from semantic_version import Version

CARGO_TREE_RE = re.compile(
    r"^[|`]-- (?P<crate>[a-z0-9_\-]+) v(?P<version>[0-9\.]+)( \(.*\))?$"
)
KOJI_RE = re.compile(
    r"^toplink/packages/rust-(?P<name>[^\/]*?)/(?P<version>[^\/]*?)/[^]*)]*"
)
VERSION_RE = re.compile(
    r"^\^(?P<major>[0-9]+)(\.(?P<minor>[0-9]+))?(\.(?P<patch>[0-9]+))?$"
)


def build_cargo_tree_dict():
    """
    Build a map of crate names to versions from the output of cargo tree.
    Determine only the versions of direct dependencies.

    :returns: a map from crates names to sets of versions
    :rtype: dict of str * set of Version
    """
    command = ["cargo", "tree", "--charset=ascii", "--all-features"]
    proc = subprocess.Popen(command, stdout=subprocess.PIPE)

    stream = proc.stdout

    version_dict = dict()
    line = stream.readline()
    while line != b"":
        line_str = line.decode("utf-8").rstrip()
        cargo_tree_match = CARGO_TREE_RE.search(line_str)
        if cargo_tree_match is not None:
            version_dict[cargo_tree_match.group("crate")] = Version(
                cargo_tree_match.group("version")
            )
        line = stream.readline()

    return version_dict


def build_koji_repo_dict(crates):
    """
    :param crates: a set of crates
    :type cargo_tree: set of str
    :returns: a dictionary containing information from the koji repo webpage
    the keys are the string representations of dependencies
    the values are the versions of dependencies
    :rtype: dict of str * Version
    """
    koji_repo_dict = {}

    requests_var = requests.get(
        "https://kojipkgs.fedoraproject.org/repos/rawhide/latest/x86_64/pkglist"
    )
    packages = requests_var.text

    for line in packages.splitlines():
        matches = KOJI_RE.match(line)
        if matches is None:
            continue
        name = matches.group("name")
        if name in crates:
            # Fedora appears to be using non-SemVer standard version strings:
            # the standard seems to be to use a "~" instead of a "-" in some
            # places. See https://semver.org/ for the canonical grammar that
            # the semantic_version library adheres to.
            version = matches.group("version").replace("~", "-")
            koji_repo_dict[name] = Version(version)

    # Post-condition: koji_repo_dict.keys() <= cargo_tree.keys().
    # cargo tree may show internal dependencies that are not separate packages
    return koji_repo_dict


def build_cargo_metadata():
    """
    Build a dict mapping crate to version spec from Cargo.toml.
    """
    command = [
        "cargo",
        "metadata",
        "--format-version=1",
        "--no-deps",
        "--all-features",
    ]
    proc = subprocess.Popen(command, stdout=subprocess.PIPE)
    stream = proc.stdout
    metadata_str = stream.readline()
    metadata = json.loads(metadata_str)
    packages = metadata["packages"]
    assert len(packages) == 1
    package = packages[0]
    dependencies = package["dependencies"]

    result = dict()
    for item in dependencies:
        matches = VERSION_RE.match(item["req"])
        major = int(matches["major"] or 0)
        minor = int(matches["minor"] or 0)
        patch = int(matches["patch"] or 0)
        result[item["name"]] = Version(major=major, minor=minor, patch=patch)

    return result



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Classes to implement dbus interface.
"""

# isort: STDLIB
import xml.etree.ElementTree as ET

# isort: FIRSTPARTY
from dbus_client_gen import managed_object_class, mo_query_builder
from dbus_python_client_gen import make_class

from ._constants import (
    BLOCKDEV_INTERFACE,
    FETCH_PROPERTIES_INTERFACE,
    FILESYSTEM_INTERFACE,
    MANAGER_INTERFACE,
    POOL_INTERFACE,
    REPORT_INTERFACE,
)
from ._introspect import SPECS

_POOL_SPEC = ET.fromstring(SPECS[POOL_INTERFACE])
_FILESYSTEM_SPEC = ET.fromstring(SPECS[FILESYSTEM_INTERFACE])
_BLOCKDEV_SPEC = ET.fromstring(SPECS[BLOCKDEV_INTERFACE])

pools = mo_query_builder(_POOL_SPEC)
filesystems = mo_query_builder(_FILESYSTEM_SPEC)
blockdevs = mo_query_builder(_BLOCKDEV_SPEC)

MOPool = managed_object_class("MOPool", _POOL_SPEC)
MOBlockDev = managed_object_class("MOBlockDev", _BLOCKDEV_SPEC)

TIME_OUT = 120  # In seconds

ObjectManager = make_class(
    "ObjectManager",
    ET.fromstring(SPECS["org.freedesktop.DBus.ObjectManager"]),
    TIME_OUT,
)
Report = make_class("Report", ET.fromstring(SPECS[REPORT_INTERFACE]), TIME_OUT)
Manager = make_class("Manager", ET.fromstring(SPECS[MANAGER_INTERFACE]), TIME_OUT)
FetchProperties = make_class(
    "FetchProperties",
    ET.fromstring(SPECS[FETCH_PROPERTIES_INTERFACE]),
    TIME_OUT,
)
Filesystem = make_class("Filesystem", _FILESYSTEM_SPEC, TIME_OUT)
Pool = make_class("Pool", _POOL_SPEC, TIME_OUT)
Blockdev = make_class("Blockdev", _BLOCKDEV_SPEC, TIME_OUT)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Top-level classes and methods.
"""

from ._connection import get_object
from ._implementation import (
    Blockdev,
    FetchProperties,
    Filesystem,
    Manager,
    MOBlockDev,
    MOPool,
    ObjectManager,
    Pool,
    Report,
    blockdevs,
    filesystems,
    pools,
)
from ._stratisd_constants import StratisdErrors



# Copyright 2021 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test that predictions of space usage match the actual.
"""

# isort: STDLIB
import json
import subprocess

# isort: LOCAL
from stratisd_client_dbus import MOPool, ObjectManager, get_object, pools
from stratisd_client_dbus._constants import TOP_OBJECT

from ._utils import (
    _STRATIS_PREDICT_USAGE,
    OptionalKeyServiceContextManager,
    ServiceContextManager,
    UdevTest,
    create_pool,
    random_string,
)


class TestSpaceUsagePrediction(UdevTest):
    """
    Test relations of prediction to reality.
    """

    _CAP_DEVICE_STR = "stratis-1-private-%s-physical-originsub"

    def _test_cap_size(self, pool_name, prediction):
        """
        Helper function to verify that the cap device is the correct size.

        :param str pool_name: the name of the pool to test
        :param prediction: JSON output from script
        """
        proxy = get_object(TOP_OBJECT)
        managed_objects = ObjectManager.Methods.GetManagedObjects(proxy, {})

        _pool_object_path, pool = next(
            pools(props={"Name": pool_name})
            .require_unique_match(True)
            .search(managed_objects)
        )

        pool_uuid = MOPool(pool).Uuid()

        cap_name = self._CAP_DEVICE_STR % pool_uuid

        command = subprocess.Popen(
            ["blockdev", "--getsize64", "/dev/mapper/%s" % cap_name],
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )
        cap_device_size, _ = command.communicate()
        self.assertEqual(cap_device_size.rstrip("\n"), prediction["free"])

    def test_prediction(self):
        """
        Verify that the prediction of space used equals the reality.
        """
        device_tokens = self._lb_mgr.create_devices(4)
        devnodes = self._lb_mgr.device_files(device_tokens)
        command = subprocess.Popen(
            [_STRATIS_PREDICT_USAGE] + devnodes, stdout=subprocess.PIPE
        )
        outs, _ = command.communicate()
        prediction = json.loads(outs)

        with ServiceContextManager():
            pool_name = random_string(5)
            create_pool(pool_name, devnodes)
            self.wait_for_pools(1)
            self._test_cap_size(pool_name, prediction)

    def test_prediction_encrypted(self):
        """
        Verify that the prediction of space used equals the reality if pool
        is encrypted.
        """
        device_tokens = self._lb_mgr.create_devices(4)
        devnodes = self._lb_mgr.device_files(device_tokens)
        command = subprocess.Popen(
            [_STRATIS_PREDICT_USAGE, "--encrypted"] + devnodes, stdout=subprocess.PIPE
        )
        outs, _ = command.communicate()
        prediction = json.loads(outs)

        (key_description, key) = ("key_spec", "data")
        with OptionalKeyServiceContextManager(key_spec=[(key_description, key)]):
            pool_name = random_string(5)
            create_pool(pool_name, devnodes, key_description=key_description)
            self.wait_for_pools(1)
            self._test_cap_size(pool_name, prediction)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Representing stratisd contants.
"""

# isort: STDLIB
from enum import Enum, IntEnum


class StratisdErrors(IntEnum):
    """
    Stratisd Errors
    """

    OK = 0
    ERROR = 1

    ALREADY_EXISTS = 2
    BUSY = 3
    INTERNAL_ERROR = 4
    NOT_FOUND = 5


class BlockDevTiers(IntEnum):
    """
    Tier to which a blockdev device belongs.
    """

    Data = 0
    Cache = 1


class EncryptionMethod(Enum):
    """
    Encryption method, used as argument to unlock.
    """

    KEYRING = "keyring"
    CLEVIS = "clevis"

    # pylint: disable=invalid-str-returned
    def __str__(self):
        return self.value



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Class to handle loop back devices.
"""

# isort: STDLIB
import os
import subprocess
import tempfile
import time
import uuid

# isort: THIRDPARTY
import pyudev

UDEV_ADD_EVENT = "add"
UDEV_REMOVE_EVENT = "remove"

_LOSETUP_BIN = os.getenv("STRATIS_LOSETUP_BIN", "/usr/sbin/losetup")

_SIZE_OF_DEVICE = 1024 ** 4  # 1 TiB


def _generate_synthetic_udev_events(devnodes, event):
    """
    Generate synthetic uevents for the given devnodes

    :param devnodes: list of device nodes
    :type devnodes: list of str
    :param str event: the event to generate, "add", "change", "remove"
    """
    for device in devnodes:
        device_name = os.path.split(device)[-1]
        ufile = os.path.join("/sys/block", device_name, "uevent")
        with open(ufile, "w") as uevent:
            uevent.write(event)


class LoopBackDevices:
    """
    Class for creating and managing loop back devices which are needed for
    specific types of udev event testing.
    """

    def __init__(self):
        """
        Class constructor which creates a temporary directory to store backing
        file in.
        """
        self.dir = tempfile.mkdtemp("_stratis_loop_back")
        self.count = 0
        self.devices = {}

    def _check_tokens(self, tokens):
        if any(token not in self.devices for token in tokens):
            raise RuntimeError("One of the specified tokens is unknown to this manager")

    def _wait_for_udev(self, tokens):
        """
        Waits for udev to detect the specified devices.

        :param tokens: identifies devices to wait for
        :type tokens: list of uuid.UUID
        :raises RuntimeError: if devices not detected in allowed time
        """
        expected_device_files = frozenset(self.device_files(tokens))

        context = pyudev.Context()
        for _ in range(10):
            if expected_device_files <= frozenset(
                [x.device_node for x in context.list_devices(subsystem="block")]
            ):
                return
            time.sleep(1)

        raise RuntimeError(
            'Loopbacked devices "%s" were created, but udev does not seem to be able to find them'
            % ", ".join(expected_device_files)
        )

    def create_devices(self, number):
        """
        Create new loop back devices.

        Note: The first time a loop back device is known it will generate
        a udev "add" event, subsequent backing file changes do not, thus we
        will need to generate it synthetically.
        :param int number: the number of devices to create
        :return: list of keys for the devices
        :rtype: list of uuid.UUID
        """
        tokens = []
        for _ in range(number):
            backing_file = os.path.join(self.dir, "block_device_%d" % self.count)
            self.count += 1

            with open(backing_file, "ab") as dev:
                dev.truncate(_SIZE_OF_DEVICE)

            device = str.strip(
                subprocess.check_output(
                    [_LOSETUP_BIN, "-f", "--show", backing_file]
                ).decode("utf-8")
            )

            token = uuid.uuid4()
            self.devices[token] = (device, backing_file)
            tokens.append(token)

        self._wait_for_udev(tokens)
        self.generate_synthetic_udev_events(tokens, UDEV_ADD_EVENT)
        return tokens

    def unplug(self, tokens):
        """
        Remove the devices from the /dev tree, but doesn't remove backing file
        :param tokens: Opaque representation of some loop back devices
        :type tokens: list of uuid.UUID
        :return: None
        :raises: RuntimeError if any token not found
        """
        self._check_tokens(tokens)
        device_files = self.device_files(tokens)

        for token in tokens:
            (device, backing_file) = self.devices[token]
            subprocess.check_call([_LOSETUP_BIN, "-d", device])
            self.devices[token] = (None, backing_file)

        _generate_synthetic_udev_events(device_files, UDEV_REMOVE_EVENT)

    def generate_synthetic_udev_events(self, tokens, event):
        """
        Synthetically create udev event for specified loop back devices
        :param tokens: Opaque representation of some loop back devices
        :type tokens: list of uuid.UUID
        :param str event: the event to generate, "add", "remove", "change"
        :return: None
        :raises RuntimeError: if any token not found or missing device node
        """
        self._check_tokens(tokens)
        _generate_synthetic_udev_events(self.device_files(tokens), event)

    def hotplug(self, tokens):
        """
        Attaches an existing backing file to specified loop back devices
        :param tokens: Opaque representation of some loop back devices
        :type tokens: list of uuid.UUID
        :return: None
        :raise RuntimeError: if token not present
        """
        self._check_tokens(tokens)
        for token in tokens:
            (_, backing_file) = self.devices[token]

            device = str.strip(
                subprocess.check_output(
                    [_LOSETUP_BIN, "-f", "--show", backing_file]
                ).decode("utf-8")
            )
            self.devices[token] = (device, backing_file)

        self.generate_synthetic_udev_events(tokens, UDEV_ADD_EVENT)

    def device_files(self, tokens):
        """
        Return the block device full names for a list of tokens
        :param tokens: Opaque representation of some loop back devices
        :type tokens: list of uuid.UUID
        :return: The list devices corresponding to the tokens
        :rtype: list of str
        :raises: KeyError if any token is not found
        :raises: RuntimeError if any devnode is None
        """
        result = [self.devices[token][0] for token in tokens]
        if any([devnode is None for devnode in result]):
            raise RuntimeError("At least one devnode is None")
        return result

    def destroy_devices(self):
        """
        Detach loopbacks and remove backing files
        :return:
        """
        for (device, backing_file) in self.devices.values():
            if device is not None:
                subprocess.check_call([_LOSETUP_BIN, "-d", device])
            os.remove(backing_file)

        self.devices = {}
        self.count = 0

    def destroy_all(self):
        """
        Detach all the devices and delete the file(s) and directory!
        :return: None
        """
        self.destroy_devices()
        os.rmdir(self.dir)
        self.dir = None
